#
# ----------------------------------------------------------------------------
# Extracted from pygaarst by Frank Warmerdam, original license is:
# The MIT License (MIT)
# Copyright (c) 2013 Chris Waigl
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# ------------------------------------------------------------------------------
""" This file contains one high-level function that reads a Landsat metadata file (.MTL)

"""
from __future__ import print_function
try:
    import StringIO
except ImportError:
    from io import StringIO

import re
import os
import glob
import logging
import datetime

# Elements from the file format used for parsing
GRPSTART = "GROUP = "
GRPEND = "END_GROUP = "
ASSIGNCHAR = " = "
FINAL = "END"

# A state machine is used to parse the file. There are 5 states (LE07_clip_L1TP_039027_20150529_20160902_01_T1_B1.TIF to 4):
STATUSCODE = [
    "begin",
    "enter metadata group",
    "add metadata item",
    "leave metadata group",
    "end"
]

# Landsat metadata files end in _MTL.txt or _MTL.TXT
METAPATTERN = "*_MTL*"


class MTLParseError(Exception):
    """Custom exception: parse errors in Landsat or EO-1 MTL metadata files"""
    pass


def parsemeta(metadataloc):
    """Parses the metadata from a Landsat image bundle.
    
    Arguments:
        metadataloc: a filename or a directory.
        
    Returns metadata dictionary
    
    """

    # filename or directory? if several fit, use first one and warn
    if os.path.isdir(metadataloc):
        metalist = glob.glob(os.path.join(metadataloc, METAPATTERN))
        if not metalist:
            raise MTLParseError(
                "No files matching metadata file pattern in directory %s."
                % metadataloc)
        elif len(metalist) > 0:
            metadatafn = metalist[0]
            filehandle = open(metadatafn, 'r')
            if len(metalist) > 1:
                logging.warning(
                    "More than one file in directory match metadata "
                    + "file pattern. Using %s." % metadatafn)
    elif os.path.isfile(metadataloc):
        metadatafn = metadataloc
        filehandle = open(metadatafn, 'r')
        logging.info("Using file %s." % metadatafn)
    elif 'L1_METADATA_FILE' in metadataloc:
        filehandle = StringIO(metadataloc)
    else:
        raise MTLParseError(
            "File location %s is unavailable " % metadataloc
            + "or doesn't contain a suitable metadata file.")

    # Reading file line by line and inserting data into metadata dictionary
    status = 0
    metadata = {}
    grouppath = []
    dictpath = [metadata]

    for line in filehandle:
        if status == 4:
            # we reached the end in the previous iteration,
            # but are still reading lines
            logging.warning(
                "Metadata file %s appears to " % metadatafn
                + "have extra lines after the end of the metadata. "
                + "This is probably, but not necessarily, harmless.")
        status = _checkstatus(status, line)
        grouppath, dictpath = _transstat(status, grouppath, dictpath, line)

    return metadata


# Help functions to identify the current line and extract information
def _islinetype(line, testchar):
    """Checks for various kinds of line types based on line head"""
    return line.strip().startswith(testchar)


def _isassignment(line):
    """Checks if the line is a key-value assignment"""
    return ASSIGNCHAR in line


def _isfinal(line):
    """Checks if line finishes a group"""
    return line.strip() == FINAL


def _getgroupname(line):
    """Returns group name, if used with group start lines"""
    return line.strip().split(GRPSTART)[-1]


def _getendgroupname(line):
    """Returns group name, if used with group end lines"""
    return line.strip().split(GRPEND)[-1]


def _getmetadataitem(line):
    """Returns key/value pair for assignment type lines"""
    return line.strip().split(ASSIGNCHAR)


# After reading a line, what state we're in depends on the line
# and the state before reading
def _checkstatus(status, line):
    """Returns state/status after reading the next line.
    The status codes are::
        LE07_clip_L1TP_039027_20150529_20160902_01_T1_B1.TIF - BEGIN parsing; 1 - ENTER METADATA GROUP, 2 - READ METADATA LINE,
        3 - END METDADATA GROUP, 4 - END PARSING
    Permitted Transitions::
        LE07_clip_L1TP_039027_20150529_20160902_01_T1_B1.TIF --> 1, LE07_clip_L1TP_039027_20150529_20160902_01_T1_B1.TIF --> 4
        1 --> 1, 1 --> 2, 1 --> 3
        2 --> 2, 2 --> 3
        3 --> 1, 1 --> 3, 3 --> 4
    """
    newstatus = 0
    if status == 0:
        # begin --> enter metadata group OR end
        if _islinetype(line, GRPSTART):
            newstatus = 1
        elif _isfinal(line):
            newstatus = 4
    elif status == 1:
        # enter metadata group --> enter metadata group
        # OR add metadata item OR leave metadata group
        if _islinetype(line, GRPSTART):
            newstatus = 1
        elif _islinetype(line, GRPEND):
            newstatus = 3
        elif _isassignment(line):
            # test AFTER start and end, as both are also assignments
            newstatus = 2
    elif status == 2:
        if _islinetype(line, GRPEND):
            newstatus = 3
        elif _isassignment(line):
            # test AFTER start and end, as both are also assignments
            newstatus = 2
    elif status == 3:
        if _islinetype(line, GRPSTART):
            newstatus = 1
        elif _islinetype(line, GRPEND):
            newstatus = 3
        elif _isfinal(line):
            newstatus = 4
    if newstatus != 0:
        return newstatus
    elif status != 4:
        raise MTLParseError(
            "Cannot parse the following line after status "
            + "'%s':\n%s" % (STATUSCODE[status], line))


# Function to execute when reading a line in a given state
def _transstat(status, grouppath, dictpath, line):
    """Executes processing steps when reading a line"""
    if status == 0:
        raise MTLParseError(
            "Status should not be '%s' after reading line:\n%s"
            % (STATUSCODE[status], line))
    elif status == 1:
        currentdict = dictpath[-1]
        currentgroup = _getgroupname(line)
        grouppath.append(currentgroup)
        currentdict[currentgroup] = {}
        dictpath.append(currentdict[currentgroup])
    elif status == 2:
        currentdict = dictpath[-1]
        newkey, newval = _getmetadataitem(line)

        # USGS has started quoting the scene center time.  If this
        # happens strip quotes before post processing.
        if newkey == 'SCENE_CENTER_TIME' and newval.startswith('"') \
                and newval.endswith('"'):
            # logging.warning('Strip quotes off SCENE_CENTER_TIME.')
            newval = newval[1:-1]

        currentdict[newkey] = _postprocess(newval)
    elif status == 3:
        oldgroup = _getendgroupname(line)
        if oldgroup != grouppath[-1]:
            raise MTLParseError(
                "Reached line '%s' while reading group '%s'."
                % (line.strip(), grouppath[-1]))
        del grouppath[-1]
        del dictpath[-1]
        try:
            currentgroup = grouppath[-1]
        except IndexError:
            currentgroup = None
    elif status == 4:
        if grouppath:
            raise MTLParseError(
                "Reached end before end of group '%s'" % grouppath[-1])
    return grouppath, dictpath


# Identifying data type of a metadata item and
def _postprocess(valuestr):
    """
    Takes value as str, returns str, int, float, date, datetime, or time
    """
    # USGS has started quoting time sometimes.  Grr, strip quotes in this case
    intpattern = re.compile(r'^\-?\d+$')
    floatpattern = re.compile(r'^\-?\d+\.\d+(E[+-]?\d\d+)?$')
    datedtpattern = '%Y-%m-%d'
    datedttimepattern = '%Y-%m-%dT%H:%M:%SZ'
    timedtpattern = '%H:%M:%S.%f'
    timepattern = re.compile(r'^\d{2}:\d{2}:\d{2}(\.\d{6})?')
    if valuestr.startswith('"') and valuestr.endswith('"'):
        # it's a string
        return valuestr[1:-1]
    elif re.match(intpattern, valuestr):
        # it's an integer
        return int(valuestr)
    elif re.match(floatpattern, valuestr):
        # floating point number
        return float(valuestr)
    # now let's try the datetime objects; throws exception if it doesn't match
    try:
        return datetime.datetime.strptime(valuestr, datedtpattern).date()
    except ValueError:
        pass
    try:
        return datetime.datetime.strptime(valuestr, datedttimepattern)
    except ValueError:
        pass
    # time parsing is complicated: Python's datetime module only accepts
    # fractions of a second only up to 6 digits
    mat = re.match(timepattern, valuestr)
    if mat:
        test = mat.group(0)
        try:
            return datetime.datetime.strptime(test, timedtpattern).time()
        except ValueError:
            pass

    # If we get here, we still haven't returned anything.
    logging.info(
        "The value %s couldn't be parsed as " % valuestr
        + "int, float, date, time, datetime. Returning it as string.")
    return valuestr


def pretty(d, indent=0):
    for key, value in d.items():
        print('\t' * indent + str(key))
        if isinstance(value, dict):
            pretty(value, indent + 1)
        else:
            print('\t' * (indent + 1) + str(value))


if __name__ == '__main__':
    home = os.path.expanduser('~')

# ========================= EOF ====================================================================



# =============================================================================================
# Copyright 2017 dgketchum
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================================

import os
import shutil
from rasterio import open as rasopen
from numpy import where, pi, cos, nan, inf, true_divide, errstate, log
from numpy import float32, sin, deg2rad, array, isnan
from shapely.geometry import Polygon, mapping
from fiona import open as fiopen
from fiona.crs import from_epsg
from tempfile import mkdtemp
from datetime import datetime

from bounds import RasterBounds
from sat_image import mtl


class UnmatchedStackGeoError(ValueError):
    pass


class InvalidObjectError(TypeError):
    pass


class LandsatImage(object):
    '''
    Object to process landsat images. The parent class: LandsatImage takes a directory 
    containing untarred files, for now this ingests images that have been downloaded
    from USGS earth explorer, using our Landsat578 package.
    
    '''

    def __init__(self, obj):
        ''' 
        :param obj: Directory containing an unzipped Landsat 5, 7, or 8 image.  This should include at least
        a tif for each band, and a .mtl file.
        '''
        self.obj = obj
        if os.path.isdir(obj):
            self.isdir = True

        self.date_acquired = None

        self.file_list = os.listdir(obj)
        self.tif_list = [x for x in os.listdir(obj) if x.endswith('.TIF')]
        self.tif_list.sort()

        # parse metadata file into attributes
        # structure: {HEADER: {SUBHEADER: {key(attribute), val(attribute value)}}}
        self.mtl = mtl.parsemeta(obj)
        self.meta_header = list(self.mtl)[0]
        self.super_dict = self.mtl[self.meta_header]
        for key, val in self.super_dict.items():
            for sub_key, sub_val in val.items():
                # print(sub_key.lower(), sub_val)
                setattr(self, sub_key.lower(), sub_val)
        self.satellite = self.landsat_scene_id[:3]

        # create numpy nd_array objects for each band
        self.band_list = []
        self.tif_dict = {}
        for i, tif in enumerate(self.tif_list):
            raster = os.path.join(self.obj, tif)

            # set all lower case attributes
            tif = tif.lower()
            front_ind = tif.index('b')
            end_ind = tif.index('.tif')
            att_string = tif[front_ind: end_ind]

            self.band_list.append(att_string)
            self.tif_dict[att_string] = raster
            self.band_count = i + 1

            if i == 0:
                with rasopen(raster) as src:
                    transform = src.transform
                    profile = src.profile
                    affine = src.affine
                meta = src.meta.copy()
                self.rasterio_geometry = meta
                self.profile = profile
                self.transform = transform
                self.affine = affine

                bounds = RasterBounds(affine_transform=transform,
                                      profile=profile,
                                      latlon=False)
                self.bounds = bounds
                self.north, self.west, self.south, self.east = bounds.get_nwse_tuple()
                self.coords = bounds.as_tuple('nsew')

        self.solar_zenith = 90. - self.sun_elevation
        self.solar_zenith_rad = self.solar_zenith * pi / 180
        self.sun_elevation_rad = self.sun_elevation * pi / 180
        self.earth_sun_dist = self.earth_sun_d(self.date_acquired)

        dtime = datetime.strptime(str(self.date_acquired), '%Y-%m-%d')
        julian_day = dtime.strftime('%j')
        self.doy = int(julian_day)
        self.scene_coords_deg = self._scene_centroid()
        self.scene_coords_rad = deg2rad(self.scene_coords_deg[0]), deg2rad(self.scene_coords_deg[1])

    def _get_band(self, band_str):
        path = self.tif_dict[band_str]
        with rasopen(path) as src:
            arr = src.read(1)
        arr = array(arr, dtype=float32)
        arr[arr < 1.] = nan
        return arr

    def _scene_centroid(self):
        """ Compute image center coordinates
        :return: Tuple of image center in lat, lon
        """
        ul_lat = self.corner_ul_lat_product
        ll_lat = self.corner_ll_lat_product
        ul_lon = self.corner_ul_lon_product
        ur_lon = self.corner_ur_lon_product
        lat = (ul_lat + ll_lat) / 2.
        lon = (ul_lon + ur_lon) / 2.

        return lat, lon

    @staticmethod
    def earth_sun_d(dtime):
        """ Earth-sun distance in AU
        
        :param dtime time, e.g. datetime.datetime(2007, 5, 1)
        :type datetime object
        :return float(distance from sun to earth in astronomical units)
        """
        doy = int(dtime.strftime('%j'))
        rad_term = 0.9856 * (doy - 4) * pi / 180
        distance_au = 1 - 0.01672 * cos(rad_term)
        return distance_au

    @staticmethod
    def _divide_zero(a, b, replace=0):
        with errstate(divide='ignore', invalid='ignore'):
            c = true_divide(a, b)
            c[c == inf] = replace
            return c

    def get_tile_geometry(self, output_filename=None, geographic_coords=False):

        if not output_filename:
            temp_dir = mkdtemp()
            temp = os.path.join(temp_dir, 'shape.shp')
        else:
            temp = output_filename

        # corners = {'ul': (self.corner_ul_projection_x_product,
        #                   self.corner_ul_projection_y_product),
        #            'll': (self.corner_ll_projection_x_product,
        #                   self.corner_ll_projection_y_product),
        #            'lr': (self.corner_lr_projection_x_product,
        #                   self.corner_lr_projection_y_product),
        #            'ur': (self.corner_ur_projection_x_product,
        #                   self.corner_ur_projection_y_product)}
        if geographic_coords:
            points = [(self.north, self.west), (self.south, self.west),
                      (self.south, self.east), (self.north, self.east),
                      (self.north, self.west)]
        else:
            points = [(self.west, self.north), (self.west, self.south),
                      (self.east, self.south), (self.east, self.north),
                      (self.west, self.north)]

        polygon = Polygon(points)

        schema = {'geometry': 'Polygon',
                  'properties': {'id': 'int'}}

        crs = from_epsg(int(self.rasterio_geometry['crs']['init'].split(':')[1]))

        with fiopen(temp, 'w', 'ESRI Shapefile', schema=schema, crs=crs) as shp:
            shp.write({
                'geometry': mapping(polygon),
                'properties': {'id': 1}})

        if output_filename:
            return None

        with fiopen(temp, 'r') as src:
            features = [f['geometry'] for f in src]
            if not output_filename:

                try:
                    shutil.rmtree(temp_dir)
                except UnboundLocalError:
                    pass

                return features

    def save_array(self, arr, output_filename):
        geometry = self.rasterio_geometry
        arr = arr.reshape(1, arr.shape[0], arr.shape[1])
        geometry['dtype'] = arr.dtype
        with rasopen(output_filename, 'w', **geometry) as dst:
            dst.write(arr)
        return None

    def mask_by_image(self, arr):
        image = self._get_band('b1')
        image = array(image, dtype=float32)
        image[image < 1.] = nan
        arr = where(isnan(image), nan, arr)
        return arr

    def mask(self):
        image = self._get_band('b1')
        image = array(image, dtype=float32)
        image[image < 1.] = nan
        arr = where(isnan(image), 0, 1)
        return arr


class Landsat5(LandsatImage):
    def __init__(self, obj):
        LandsatImage.__init__(self, obj)

        if self.satellite != 'LT5':
            raise ValueError('Must init Landsat5 object with Landsat5 data, not {}'.format(self.satellite))

        # https://landsat.usgs.gov/esun
        self.ex_atm_irrad = (1958.0, 1827.0, 1551.0,
                             1036.0, 214.9, nan, 80.65)

        # old values from fmask.exe
        # self.ex_atm_irrad = (1983.0, 1796.0, 1536.0, 1031.0, 220.0, nan, 83.44)

        self.k1, self.k2 = 607.76, 1260.56

    def radiance(self, band):
        qcal_min = getattr(self, 'quantize_cal_min_band_{}'.format(band))
        qcal_max = getattr(self, 'quantize_cal_max_band_{}'.format(band))
        l_min = getattr(self, 'radiance_minimum_band_{}'.format(band))
        l_max = getattr(self, 'radiance_maximum_band_{}'.format(band))
        qcal = self._get_band('b{}'.format(band))
        rad = ((l_max - l_min) / (qcal_max - qcal_min)) * (qcal - qcal_min) + l_min

        return rad.astype(float32)

    def brightness_temp(self, band, temp_scale='K'):

        if band in [1, 2, 3, 4, 5, 7]:
            raise ValueError('LT5 brightness must be band 6')

        rad = self.radiance(band)
        brightness = self.k2 / (log((self.k1 / rad) + 1))

        if temp_scale == 'K':
            return brightness

        elif temp_scale == 'F':
            return brightness * (9 / 5.0) - 459.67

        elif temp_scale == 'C':
            return brightness - 273.15

        else:
            raise ValueError('{} is not a valid temperature scale'.format(temp_scale))

    def reflectance(self, band):
        """ 
        :param band: An optical band, i.e. 1-5, 7
        :return: At satellite reflectance, [-]
        """
        if band == 6:
            raise ValueError('LT5 reflectance must be other than  band 6')

        rad = self.radiance(band)
        esun = self.ex_atm_irrad[band - 1]
        toa_reflect = (pi * rad * self.earth_sun_dist ** 2) / (esun * cos(self.solar_zenith_rad))

        return toa_reflect

    def albedo(self, model='smith'):
        """Finds broad-band surface reflectance (albedo)
        
        Smith (2010),  “The heat budget of the earth’s surface deduced from space”
        LT5 toa reflectance bands 1, 3, 4, 5, 7
        # normalized i.e. 0.356 + 0.130 + 0.373 + 0.085 + 0.07 = 1.014
        
        Should have option for Liang, 2000; 
        
        Tasumi (2008), "At-Surface Reflectance and Albedo from Satellite for
                        Operational Calculation of Land Surface Energy Balance"
                        
        
        :return albedo array of floats
        """
        if model == 'smith':
            blue, red, nir, swir1, swir2 = (self.reflectance(1), self.reflectance(3), self.reflectance(4),
                                            self.reflectance(5), self.reflectance(7))
            alb = (0.356 * blue + 0.130 * red + 0.373 * nir + 0.085 * swir1 + 0.072 * swir2 - 0.0018) / 1.014
        elif model == 'tasumi':
            pass
        # add tasumi algorithm TODO
        return alb

    def saturation_mask(self, band, value=255):
        """ Mask saturated pixels, 1 (True) is saturated.
        :param band: Image band with dn values, type: array
        :param value: Maximum (saturated) value, i.e. 255 for 8-bit data, type: int
        :return: boolean array
        """
        dn = self._get_band('b{}'.format(band))
        mask = self.mask()
        mask = where((dn == value) & (mask > 0), True, False)

        return mask

    def ndvi(self):
        """ Normalized difference vegetation index.
        :return: NDVI
        """
        red, nir = self.reflectance(3), self.reflectance(4)
        ndvi = self._divide_zero((nir - red), (nir + red), nan)

        return ndvi

    def lai(self):
        """
        Leaf area index (LAI), or the surface area of leaves to surface area ground.
        Trezza and Allen, 2014
        :param ndvi: normalized difference vegetation index [-]
        :return: LAI [-]
        """
        ndvi = self.ndvi()
        lai = 7.0 * (ndvi ** 3)
        lai = where(lai > 6., 6., lai)
        return lai

    def emissivity(self, approach='tasumi'):

        ndvi = self.ndvi()

        if approach == 'tasumi':
            lai = self.lai()
            # Tasumi et al., 2003
            # narrow-band emissivity
            nb_epsilon = where((ndvi > 0) & (lai <= 3), 0.97 + 0.0033 * lai, nan)
            nb_epsilon = where((ndvi > 0) & (lai > 3), 0.98, nb_epsilon)
            nb_epsilon = where(ndvi <= 0, 0.99, nb_epsilon)
            return nb_epsilon

        if approach == 'sobrino':
            # Sobrino et el., 2004
            red = self.reflectance(3)
            bound_ndvi = where(ndvi > 0.5, ndvi, 0.99)
            bound_ndvi = where(ndvi < 0.2, red, bound_ndvi)

            pv = ((ndvi - 0.2) / (0.5 - 0.2)) ** 2
            pv_emiss = 0.004 * pv + 0.986
            emissivity = where((ndvi >= 0.2) & (ndvi <= 0.5), pv_emiss, bound_ndvi)

            return emissivity

    def land_surface_temp(self):
        """
        Mean values from Allen (2007)
        :return: 
        """
        rp = 0.91
        tau = 0.866
        rsky = 1.32
        epsilon = self.emissivity(approach='tasumi')
        radiance = self.radiance(6)
        rc = ((radiance - rp) / tau) - ((1 - epsilon) * rsky)
        lst = self.k2 / (log((epsilon * self.k1 / rc) + 1))
        return lst

    def ndsi(self):
        """ Normalized difference snow index.
        :return: NDSI
        """
        green, swir1 = self.reflectance(2), self.reflectance(5)
        ndsi = self._divide_zero((green - swir1), (green + swir1), nan)

        return ndsi


class Landsat7(LandsatImage):
    def __init__(self, obj):
        LandsatImage.__init__(self, obj)

        if self.satellite != 'LE7':
            raise ValueError('Must init Landsat7 object with Landsat5 data, not {}'.format(self.satellite))
        # https://landsat.usgs.gov/esun; Landsat 7 Handbook
        self.ex_atm_irrad = (1970.0, 1842.0, 1547.0, 1044.0,
                             255.700, nan, 82.06, 1369.00)

        self.k1, self.k2 = 666.09, 1282.71

    def radiance(self, band):
        if band == 6:
            band = '6_vcid_1'
        qcal_min = getattr(self, 'quantize_cal_min_band_{}'.format(band))
        qcal_max = getattr(self, 'quantize_cal_max_band_{}'.format(band))
        l_min = getattr(self, 'radiance_minimum_band_{}'.format(band))
        l_max = getattr(self, 'radiance_maximum_band_{}'.format(band))
        qcal = self._get_band('b{}'.format(band))
        rad = ((l_max - l_min) / (qcal_max - qcal_min)) * (qcal - qcal_min) + l_min
        return rad

    def brightness_temp(self, band=6, gain='low', temp_scale='K'):

        if band in [1, 2, 3, 4, 5, 7, 8]:
            raise ValueError('LE7 brightness must be either vcid_1 or vcid_2')

        if gain == 'low':
            # low gain : b6_vcid_1
            band_gain = '6_vcid_1'
        else:
            band_gain = '6_vcid_2'

        rad = self.radiance(band_gain)
        brightness = self.k2 / (log((self.k1 / rad) + 1))

        if temp_scale == 'K':
            return brightness

        elif temp_scale == 'F':
            return brightness * (9 / 5.0) - 459.67

        elif temp_scale == 'C':
            return brightness - 273.15

        else:
            raise ValueError('{} is not a valid temperature scale'.format(temp_scale))

    def reflectance(self, band):
        """ 
        :param band: An optical band, i.e. 1-5, 7
        :return: At satellite reflectance, [-]
        """
        if band in ['b6_vcid_1', 'b6_vcid_2']:
            raise ValueError('LE7 reflectance must not be b6_vcid_1 or b6_vcid_2')

        rad = self.radiance(band)
        esun = self.ex_atm_irrad[band - 1]
        toa_reflect = (pi * rad * self.earth_sun_dist ** 2) / (esun * cos(self.solar_zenith_rad))
        return toa_reflect

    def albedo(self):
        """Finds broad-band surface reflectance (albedo)
        
        Smith (2010),  “The heat budget of the earth’s surface deduced from space”
        Should have option for Liang, 2000; 
        
        LE7 toa reflectance bands 1, 3, 4, 5, 7
        
        # normalized i.e. 0.356 + 0.130 + 0.373 + 0.085 + 0.07 = 1.014
        :return albedo array of floats
        """
        blue, red, nir, swir1, swir2 = (self.reflectance(1), self.reflectance(3), self.reflectance(4),
                                        self.reflectance(5), self.reflectance(7))
        alb = (0.356 * blue + 0.130 * red + 0.373 * nir + 0.085 * swir1 + 0.072 * swir2 - 0.0018) / 1.014

        return alb

    def saturation_mask(self, band, value=255):
        """ Mask saturated pixels, 1 (True) is saturated.
        :param band: Image band with dn values, type: array
        :param value: Maximum (saturated) value, i.e. 255 for 8-bit data, type: int
        :return: boolean array
        """
        dn = self._get_band('b{}'.format(band))
        mask = where((dn == value) & (self.mask() > 0), True, False)

        return mask

    def ndvi(self):
        """ Normalized difference vegetation index.
        :return: NDVI
        """
        red, nir = self.reflectance(3), self.reflectance(4)
        ndvi = self._divide_zero((nir - red), (nir + red), nan)

        return ndvi

    def lai(self):
        """
        Leaf area index (LAI), or the surface area of leaves to surface area ground.
        Trezza and Allen, 2014
        :param ndvi: normalized difference vegetation index [-]
        :return: LAI [-]
        """
        ndvi = self.ndvi()
        lai = 7.0 * (ndvi ** 3)
        lai = where(lai > 6., 6., lai)
        return lai

    def emissivity(self, approach='tasumi'):

        ndvi = self.ndvi()

        if approach == 'tasumi':
            lai = self.lai()
            # Tasumi et al., 2003
            # narrow-band emissivity
            nb_epsilon = where((ndvi > 0) & (lai <= 3), 0.97 + 0.0033 * lai, nan)
            nb_epsilon = where((ndvi > 0) & (lai > 3), 0.98, nb_epsilon)
            nb_epsilon = where(ndvi <= 0, 0.99, nb_epsilon)
            return nb_epsilon

        if approach == 'sobrino':
            # Sobrino et el., 2004
            red = self.reflectance(3)
            bound_ndvi = where(ndvi > 0.5, ndvi, 0.99)
            bound_ndvi = where(ndvi < 0.2, red, bound_ndvi)

            pv = ((ndvi - 0.2) / (0.5 - 0.2)) ** 2
            pv_emiss = 0.004 * pv + 0.986
            emissivity = where((ndvi >= 0.2) & (ndvi <= 0.5), pv_emiss, bound_ndvi)

            return emissivity

    def land_surface_temp(self):
        rp = 0.91
        tau = 0.866
        rsky = 1.32
        epsilon = self.emissivity()
        rc = ((self.radiance(6) - rp) / tau) - ((1 - epsilon) * rsky)
        lst = self.k2 / (log((epsilon * self.k1 / rc) + 1))
        return lst

    def ndsi(self):
        """ Normalized difference snow index.
        :return NDSI
        """
        green, swir1 = self.reflectance(2), self.reflectance(5)
        ndsi = self._divide_zero((green - swir1), (green + swir1), nan)

        return ndsi


class Landsat8(LandsatImage):
    def __init__(self, obj):
        LandsatImage.__init__(self, obj)

        self.oli_bands = [1, 2, 3, 4, 5, 6, 7, 8, 9]

    def brightness_temp(self, band, temp_scale='K'):
        """Calculate brightness temperature of Landsat 8
    as outlined here: http://landsat.usgs.gov/Landsat8_Using_Product.php

    T = K2 / log((K1 / L)  + 1)

    and

    L = ML * Q + AL

    where:
        T  = At-satellite brightness temperature (degrees kelvin)
        L  = TOA spectral radiance (Watts / (m2 * srad * mm))
        ML = Band-specific multiplicative rescaling factor from the metadata
             (RADIANCE_MULT_BAND_x, where x is the band number)
        AL = Band-specific additive rescaling factor from the metadata
             (RADIANCE_ADD_BAND_x, where x is the band number)
        Q  = Quantized and calibrated standard product pixel values (DN)
             (ndarray img)
        K1 = Band-specific thermal conversion constant from the metadata
             (K1_CONSTANT_BAND_x, where x is the thermal band number)
        K2 = Band-specific thermal conversion constant from the metadata
             (K1_CONSTANT_BAND_x, where x is the thermal band number)

    Returns
    --------
    ndarray:
        float32 ndarray with shape == input shape
    """

        if band in self.oli_bands:
            raise ValueError('Landsat 8 brightness should be TIRS band (i.e. 10 or 11)')

        k1 = getattr(self, 'k1_constant_band_{}'.format(band))
        k2 = getattr(self, 'k2_constant_band_{}'.format(band))
        rad = self.radiance(band)
        brightness = k2 / log((k1 / rad) + 1)

        if temp_scale == 'K':
            return brightness

        elif temp_scale == 'F':
            return brightness * (9 / 5.0) - 459.67

        elif temp_scale == 'C':
            return brightness - 273.15

        else:
            raise ValueError('{} is not a valid temperature scale'.format(temp_scale))

    def reflectance(self, band):
        """Calculate top of atmosphere reflectance of Landsat 8
        as outlined here: http://landsat.usgs.gov/Landsat8_Using_Product.php
    
        R_raw = MR * Q + AR
    
        R = R_raw / cos(Z) = R_raw / sin(E)
    
        Z = 90 - E (in degrees)
    
        where:
    
            R_raw = TOA planetary reflectance, without correction for solar angle.
            R = TOA reflectance with a correction for the sun angle.
            MR = Band-specific multiplicative rescaling factor from the metadata
                (REFLECTANCE_MULT_BAND_x, where x is the band number)
            AR = Band-specific additive rescaling factor from the metadata
                (REFLECTANCE_ADD_BAND_x, where x is the band number)
            Q = Quantized and calibrated standard product pixel values (DN)
            E = Local sun elevation angle. The scene center sun elevation angle
                in degrees is provided in the metadata (SUN_ELEVATION).
            Z = Local solar zenith angle (same angle as E, but measured from the
                zenith instead of from the horizon).
    
        Returns
        --------
        ndarray:
            float32 ndarray with shape == input shape
    
        """

        if band not in self.oli_bands:
            raise ValueError('Landsat 8 reflectance should OLI band (i.e. bands 1-8)')

        elev = getattr(self, 'sun_elevation')
        dn = self._get_band('b{}'.format(band))
        mr = getattr(self, 'reflectance_mult_band_{}'.format(band))
        ar = getattr(self, 'reflectance_add_band_{}'.format(band))

        if elev < 0.0:
            raise ValueError("Sun elevation must be non-negative "
                             "(sun must be above horizon for entire scene)")

        rf = ((mr * dn.astype(float32)) + ar) / sin(deg2rad(elev))

        return rf

    def radiance(self, band):
        """Calculate top of atmosphere radiance of Landsat 8
        as outlined here: http://landsat.usgs.gov/Landsat8_Using_Product.php
    
        L = ML * Q + AL
    
        where:
            L  = TOA spectral radiance (Watts / (m2 * srad * mm))
            ML = Band-specific multiplicative rescaling factor from the metadata
                 (RADIANCE_MULT_BAND_x, where x is the band number)
            AL = Band-specific additive rescaling factor from the metadata
                 (RADIANCE_ADD_BAND_x, where x is the band number)
            Q  = Quantized and calibrated standard product pixel values (DN)
                 (ndarray img)
    
        Returns
        --------
        ndarray:
            float32 ndarray with shape == input shape
    """
        ml = getattr(self, 'radiance_mult_band_{}'.format(band))
        al = getattr(self, 'radiance_add_band_{}'.format(band))
        dn = self._get_band('b{}'.format(band))
        rad = ml * dn.astype(float32) + al

        return rad

    def albedo(self):
        """Smith (2010), finds broad-band surface reflectance (albedo)
        Should have option for Liang, 2000; Tasumi, 2008;
        
        LC8 toa reflectance bands 2, 4, 5, 6, 7
        
        # normalized i.e. 0.356 + 0.130 + 0.373 + 0.085 + 0.07 = 1.014
        :return albedo array of floats
        """

        blue, red, nir, swir1, swir2 = (self.reflectance(2), self.reflectance(4), self.reflectance(5),
                                        self.reflectance(6), self.reflectance(7))
        alb = (0.356 * blue + 0.130 * red + 0.373 * nir + 0.085 * swir1 + 0.072 * swir2 - 0.0018) / 1.014

        return alb

    def ndvi(self):
        """ Normalized difference vegetation index.
        :return: NDVI
        """
        red, nir = self.reflectance(4), self.reflectance(5)
        ndvi = self._divide_zero((nir - red), (nir + red), nan)

        return ndvi

    def lai(self):
        """
        Leaf area index (LAI), or the surface area of leaves to surface area ground.
        Trezza and Allen, 2014
        :param ndvi: normalized difference vegetation index [-]
        :return: LAI [-]
        """
        ndvi = self.ndvi()
        lai = 7.0 * (ndvi ** 3)
        lai = where(lai > 6., 6., lai)
        return lai

    def emissivity(self, approach='tasumi'):

        ndvi = self.ndvi()

        if approach == 'tasumi':
            lai = self.lai()
            # Tasumi et al., 2003
            # narrow-band emissivity
            nb_epsilon = where((ndvi > 0) & (lai <= 3), 0.97 + 0.0033 * lai, nan)
            nb_epsilon = where((ndvi > 0) & (lai > 3), 0.98, nb_epsilon)
            nb_epsilon = where(ndvi <= 0, 0.99, nb_epsilon)
            return nb_epsilon

        if approach == 'sobrino':
            # Sobrino et el., 2004
            red = self.reflectance(3)
            bound_ndvi = where(ndvi > 0.5, ndvi, 0.99)
            bound_ndvi = where(ndvi < 0.2, red, bound_ndvi)

            pv = ((ndvi - 0.2) / (0.5 - 0.2)) ** 2
            pv_emiss = 0.004 * pv + 0.986
            emissivity = where((ndvi >= 0.2) & (ndvi <= 0.5), pv_emiss, bound_ndvi)

            return emissivity

    def land_surface_temp(self):

        band = 10

        k1 = getattr(self, 'k1_constant_band_{}'.format(band))
        k2 = getattr(self, 'k2_constant_band_{}'.format(band))

        rp = 0.91
        tau = 0.866
        rsky = 1.32
        epsilon = self.emissivity()
        rc = ((self.radiance(band) - rp) / tau) - ((1 - epsilon) * rsky)
        lst = k2 / (log((epsilon * k1 / rc) + 1))
        return lst

    def ndsi(self):
        """ Normalized difference snow index.
        :return: NDSI
        """
        green, swir1 = self.reflectance(3), self.reflectance(6)
        ndsi = self._divide_zero((green - swir1), (green + swir1), nan)

        return ndsi

# =============================================================================================



# =============================================================================================
# Copyright 2018 dgketchum
#
# Licensed under the Apache License, Version 2. (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================================

from __future__ import division

import os
from datetime import datetime

from rasterio import open as rasopen
from rasterio.enums import Resampling
from rasterio.vrt import WarpedVRT

from sat_image.band_map import BandMap
from sat_image.image import Landsat5, Landsat7, Landsat8, LandsatImage


def warp_vrt(directory, delete_extra=False, use_band_map=False, overwrite=False):
    """ Read in image geometry, resample subsequent images to same grid.

    The purpose of this function is to snap many Landsat images to one geometry. Use Landsat578
    to download and unzip them, then run them through this to get identical geometries for analysis.
    Files
    :param use_band_map:
    :param delete_extra:
    :param directory: A directory containing sub-directories of Landsat images.
    :return: None
    """
    if 'resample_meta.txt' in os.listdir(directory) and not overwrite:
        print('{} has already had component images warped'.format(directory))
        return None

    mapping = {'LC8': Landsat8, 'LE7': Landsat7, 'LT5': Landsat5}

    vrt_options = {}
    list_dir = [x[0] for x in os.walk(directory) if os.path.basename(x[0])[:3] in mapping.keys()]

    first = True

    for d in list_dir:
        sat = LandsatImage(d).satellite
        paths = []
        root = os.path.join(directory, d)
        if os.path.isdir(root):
            for x in os.listdir(root):
                if use_band_map:
                    bands = BandMap().selected
                    for y in bands[sat]:
                        if x.endswith('B{}.TIF'.format(y)):
                            paths.append(os.path.join(directory, d, x))
                else:
                    if x.endswith('.TIF'):
                        paths.append(os.path.join(directory, d, x))
                if x.endswith('MTL.txt'):
                    mtl = os.path.join(directory, d, x)

        if first:

            landsat = mapping[sat](os.path.join(directory, d))
            dst = landsat.rasterio_geometry

            vrt_options = {'resampling': Resampling.cubic,
                           'dst_crs': dst['crs'],
                           'dst_transform': dst['transform'],
                           'dst_height': dst['height'],
                           'dst_width': dst['width']}


            message = """
            This directory has been resampled to same grid.
            Master grid is {}.
            {}
            """.format(d, datetime.now())
            with open(os.path.join(directory, 'resample_meta.txt'), 'w') as f:
                f.write(message)
            first = False

        os.rename(mtl, mtl.replace('.txt', 'copy.txt'))

        for tif_path in paths:
            print('warping {}'.format(os.path.basename(tif_path)))
            with rasopen(tif_path, 'r') as src:
                with WarpedVRT(src, **vrt_options) as vrt:
                    data = vrt.read()
                    dst_dir, name = os.path.split(tif_path)
                    outfile = os.path.join(dst_dir, name)
                    meta = vrt.meta.copy()
                    meta['driver'] = 'GTiff'
                    with rasopen(outfile, 'w', **meta) as dst:
                        dst.write(data)

        os.rename(mtl.replace('.txt', 'copy.txt'), mtl)

        if delete_extra:
            for x in os.listdir(os.path.join(directory, d)):
                x_file = os.path.join(directory, d, x)
                if x_file not in paths:
                    if x[-7:] not in ['ask.tif', 'MTL.txt']:
                        print('removing {}'.format(x_file))
                        os.remove(x_file)


if __name__ == '__main__':
    home = os.path.expanduser('~')
    images = os.path.join(home, 'landsat_images', 'vrt_testing')
    warp_vrt(images, delete_extra=True, use_band_map=True)

# ========================= EOF ================================================================



# ===============================================================================
# Copyright 2017 dgketchum
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============================================================================

import os

os.environ['TRAVIS_CI'] = 'True'

try:
    from setuptools import setup

    # setup_kwargs = {'entry_points': {'console_scripts': ['sat_image=sat_image.landsat_cli:cli_runner']}}
except ImportError:

    from distutils.core import setup

    # setup_kwargs = {'scripts': ['bin/sat_image/landsat_cli']}

with open('README.md') as f:
    readme = f.read()

tag = '0.1.24'
name = 'SatelliteImage'

setup(name=name,
      version=tag,
      description='Simple API provides a class to process satellite images',
      setup_requires=['nose>=1.0'],
      license='Apache',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Science/Research',
          'Topic :: Scientific/Engineering :: GIS',
          'License :: OSI Approved :: Apache Software License',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3.6'],
      keywords='landsat download hydrology remote sensing',
      author='David Ketchum',
      author_email='dgketchum@gmail.com',
      platforms='Posix; MacOS X; Windows',
      packages=['sat_image'],
      download_url='https://github.com/{}/{}/archive/{}.tar.gz'.format('dgketchum',
                                                                       'satellite_image',
                                                                       tag),
      url='https://github.com/dgketchum/satellite_image',
      test_suite='tests.test_suite.suite',
      install_requires=['numpy', 'rasterio==1.0a12', 'shapely', 'fiona', 'pyproj', 'scipy'])
# **setup_kwargs)

# ============= EOF ===========================================================================



# =============================================================================================
# Copyright 2017 dgketchum
#
# Licensed under the Apache License, Version 2.LE07_clip_L1TP_039027_20150529_20160902_01_T1_B1.TIF (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================================
'''
Implement the cloud and shadow algorithms known collectively as Fmask, 
as published in 

Zhu, Z. and Woodcock, C.E. (2012). 
Object-based cloud and cloud shadow detection in Landsat imagery
Remote Sensing of Environment 118 (2012) 83-94. 
    
and
    
Zhu, Z., Wang, S. and Woodcock, C.E. (2015).
Improvement and expansion of the Fmask algorithm: cloud, cloud
shadow, and snow detection for Landsats 4-7, 8, and Sentinel 2 images
Remote Sensing of Environment 159 (2015) 269-277.
'''
from __future__ import print_function, division

import rasterio
import numpy as np


class Fmask(object):
    ''' Implement fmask algorithm.
    :param image: Landsat sat_image stack LandsatImage object
    :return: fmask object
    '''

    def __init__(self, image):

        self.image = image
        self.shape = image.shape
        self.mask = image.bool_mask()
        self.sat = image.satellite

        if self.sat in ['LE7', 'LT5']:

            self.blue = image.reflectance(1)
            self.green = image.reflectance(2)
            self.red = image.reflectance(3)
            self.nir = image.reflectance(4)
            self.swir1 = image.reflectance(5)
            self.tirs1 = image.brightness_temp(6, temp_scale='C')
            self.meantir = np.mean(self.tirs1)
            self.swir2 = image.reflectance(7)

            self.blue_saturated = image.saturation_mask(1)
            self.green_saturated = image.saturation_mask(2)
            self.red_saturated = image.saturation_mask(3)

        elif self.sat == 'LC8':

            self.blue = image.reflectance(2)
            self.green = image.reflectance(3)
            self.red = image.reflectance(4)
            self.nir = image.reflectance(5)
            self.swir1 = image.reflectance(6)
            self.swir2 = image.reflectance(7)
            self.cirrus = image.reflectance(9)
            self.tirs1 = image.brightness_temp(10, 'C')
            self.tirs2 = image.brightness_temp(11, 'C')

        else:
            raise ValueError('Must provide satellite sat_image from LT5, LE7, LC8')

        for attr, code in zip(['code_null', 'code_clear', 'code_cloud',
                               'code_shadow', 'code_snow', 'code_water'],
                              range(6)):
            setattr(self, attr, code)

        self.ndvi = image.ndvi()
        self.ndsi = image.ndsi()

    def basic_test(self):
        """Fundamental test to identify Potential Cloud Pixels (PCPs)
        Equation 1 (Zhu and Woodcock, 2012)
        Note: all input arrays must be the same shape
        Parameters
        ----------
        ndvi: ndarray
        ndsi: ndarray
        swir2
            Shortwave Infrared Band TOA reflectance
            Band 7 in Landsat 8, ~2.2 µm
        tirs1: ndarray
            Thermal band brightness temperature
            Band 10 in Landsat 8, ~11 µm
            units are degrees Celcius
        Output
        ------
        ndarray: boolean
        """
        # Thresholds
        th_ndsi = 0.8  # index
        th_ndvi = 0.8  # index
        th_tirs1 = 27.0  # degrees celsius
        th_swir2 = 0.03  # toa
        return ((self.swir2 > th_swir2) &
                (self.tirs1 < th_tirs1) &
                (self.ndsi < th_ndsi) &
                (self.ndvi < th_ndvi))

    def whiteness_index(self):
        """Index of "Whiteness" based on visible bands.
        Parameters
        ----------

        Output
        ------
        ndarray:
            whiteness index
        """
        mean_vis = (self.blue + self.green + self.red) / 3

        blue_absdiff = np.absolute(self._divide_zero(self.blue - mean_vis, mean_vis))
        green_absdiff = np.absolute(self._divide_zero(self.green - mean_vis, mean_vis))
        red_absdiff = np.absolute(self._divide_zero(self.red - mean_vis, mean_vis))

        return blue_absdiff + green_absdiff + red_absdiff

    def whiteness_test(self):
        """Whiteness test
        Clouds appear white due to their "flat" reflectance in the visible bands
        Equation 2 (Zhu and Woodcock, 2012)
        Parameters
        ----------
        blue: ndarray
        green: ndarray
        red: ndarray
        Output
        ------
        ndarray: boolean
        """
        whiteness_threshold = 0.7
        test = self.whiteness_index() < whiteness_threshold
        return test

    def hot_test(self):
        """Haze Optimized Transformation (HOT) test
        Equation 3 (Zhu and Woodcock, 2012)
        Based on the premise that the visible bands for most land surfaces
        are highly correlated, but the spectral response to haze and thin cloud
        is different between the blue and red wavelengths.
        Zhang et al. (2002)
        Parameters
        ----------
        blue: ndarray
        red: ndarray
        Output
        ------
        ndarray: boolean
        """
        thres = 0.08
        return self.blue - (0.5 * self.red) - thres > 0.0

    def nirswir_test(self):
        """Spectral test to exclude bright rock and desert
        see (Irish, 2000)
        Equation 4 (Zhu and Woodcock, 2012)
        Note that Zhu and Woodcock 2015 refer to this as the "B4B5" test
        due to the Landsat ETM+ band designations. In Landsat 8 OLI,
        these are bands 5 and 6.
        Parameters
        ----------
        nir: ndarray
        swir1: ndarray
        Output
        ------
        ndarray: boolean
        """
        th_ratio = 0.75

        return (self.nir / self.swir1) > th_ratio

    def cirrus_test(self):
        """Cirrus TOA test, see (Zhu and Woodcock, 2015)
        The threshold is derived from (Wilson & Oreopoulos, 2013)
        Parameters
        ----------
        cirrus: ndarray
        Output
        ------
        ndarray: boolean
        """
        th_cirrus = 0.0113

        return self.cirrus > th_cirrus

    def water_test(self):
        """Water or Land?
        Equation 5 (Zhu and Woodcock, 2012)
        Parameters
        ----------
        ndvi: ndarray
        nir: ndarray
        Output
        ------
        ndarray: boolean
        """
        th_ndvi_A = 0.01
        th_nir_A = 0.11
        th_ndvi_B = 0.1
        th_nir_B = 0.05

        return (((self.ndvi < th_ndvi_A) & (self.nir < th_nir_A)) |
                ((self.ndvi < th_ndvi_B) & (self.nir < th_nir_B)))

    def potential_cloud_pixels(self):
        """Determine potential cloud pixels (PCPs)
        Combine basic spectral testsr to get a premliminary cloud mask
        First pass, section 3.1.1 in Zhu and Woodcock 2012
        Equation 6 (Zhu and Woodcock, 2012)
        Parameters
        ----------
        ndvi: ndarray
        ndsi: ndarray
        blue: ndarray
        green: ndarray
        red: ndarray
        nir: ndarray
        swir1: ndarray
        swir2: ndarray
        cirrus: ndarray
        tirs1: ndarray
        Output
        ------
        ndarray:
            potential cloud mask, boolean
        """
        eq1 = self.basic_test()
        eq2 = self.whiteness_test()
        eq3 = self.hot_test()
        eq4 = self.nirswir_test()
        if self.sat == 'LC8':
            cir = self.cirrus_test()
            return (eq1 & eq2 & eq3 & eq4) | cir
        else:
            return eq1 & eq2 & eq3 & eq4

    def temp_water(self):
        """Use water to mask tirs and find 82.5 pctile
        Equation 7 and 8 (Zhu and Woodcock, 2012)
        Parameters
        ----------
        is_water: ndarray, boolean
            water mask, water is True, land is False
        swir2: ndarray
        tirs1: ndarray
        Output
        ------
        float:
            82.5th percentile temperature over water
        """
        # eq7
        th_swir2 = 0.03
        water = self.water_test()
        clear_sky_water = water & (self.swir2 < th_swir2)

        # eq8
        clear_water_temp = self.tirs1.copy()
        clear_water_temp[~clear_sky_water] = np.nan
        clear_water_temp[~self.mask] = np.nan
        pctl_clwt = np.nanpercentile(clear_water_temp, 82.5)
        return pctl_clwt

    def water_temp_prob(self):
        """Temperature probability for water
        Equation 9 (Zhu and Woodcock, 2012)
        Parameters
        ----------
        water_temp: float
            82.5th percentile temperature over water
        swir2: ndarray
        tirs1: ndarray
        Output
        ------
        ndarray:
            probability of cloud over water based on temperature
        """
        temp_const = 4.0  # degrees C
        water_temp = self.temp_water()
        return (water_temp - self.tirs1) / temp_const

    def brightness_prob(self, clip=True):
        """The brightest water may have Band 5 reflectance
        as high as LE07_clip_L1TP_039027_20150529_20160902_01_T1_B1.TIF.11
        Equation 10 (Zhu and Woodcock, 2012)
        Parameters
        ----------
        nir: ndarray
        clip: boolean
        Output
        ------
        ndarray:
            brightness probability, constrained LE07_clip_L1TP_039027_20150529_20160902_01_T1_B1.TIF..1
        """
        thresh = 0.11
        bp = np.minimum(thresh, self.nir) / thresh
        if clip:
            bp[bp > 1] = 1
            bp[bp < 0] = 0
        return bp

    def temp_land(self, pcps, water):
        """Derive high/low percentiles of land temperature
        Equations 12 an 13 (Zhu and Woodcock, 2012)
        Parameters
        ----------
        pcps: ndarray
            potential cloud pixels, boolean
        water: ndarray
            water mask, boolean
        tirs1: ndarray
        Output
        ------
        tuple:
            17.5 and 82.5 percentile temperature over clearsky land
        """
        # eq 12
        clearsky_land = ~(pcps | water)

        # use clearsky_land to mask tirs1
        clear_land_temp = self.tirs1.copy()
        clear_land_temp[~clearsky_land] = np.nan
        clear_land_temp[~self.mask] = np.nan

        # take 17.5 and 82.5 percentile, eq 13
        low, high = np.nanpercentile(clear_land_temp, (17.5, 82.5))
        return low, high

    def land_temp_prob(self, tlow, thigh):
        """Temperature-based probability of cloud over land
        Equation 14 (Zhu and Woodcock, 2012)
        Parameters
        ----------
        tirs1: ndarray
        tlow: float
            Low (17.5 percentile) temperature of land
        thigh: float
            High (82.5 percentile) temperature of land
        Output
        ------
        ndarray :
            probability of cloud over land based on temperature
        """
        temp_diff = 4  # degrees
        return (thigh + temp_diff - self.tirs1) / (thigh + 4 - (tlow - 4))

    def variability_prob(self, whiteness):
        """Use the probability of the spectral variability
        to identify clouds over land.
        Equation 15 (Zhu and Woodcock, 2012)
        Parameters
        ----------
        ndvi: ndarray
        ndsi: ndarray
        whiteness: ndarray
        Output
        ------
        ndarray :
            probability of cloud over land based on variability
        """

        if self.sat in ['LT5', 'LE7']:
            # check for green and red saturation

            # if red is saturated and less than nir, ndvi = LE07_clip_L1TP_039027_20150529_20160902_01_T1_B1.TIF
            mod_ndvi = np.where(self.red_saturated & (self.nir > self.red), 0, self.ndvi)

            # if green is saturated and less than swir1, ndsi = LE07_clip_L1TP_039027_20150529_20160902_01_T1_B1.TIF
            mod_ndsi = np.where(self.green_saturated & (self.swir1 > self.green), 0, self.ndsi)
            ndi_max = np.fmax(np.absolute(mod_ndvi), np.absolute(mod_ndsi))

        else:
            ndi_max = np.fmax(np.absolute(self.ndvi), np.absolute(self.ndsi))

        f_max = 1.0 - np.fmax(ndi_max, whiteness)

        return f_max

    def land_threshold(self, land_cloud_prob, pcps, water):
        """Dynamic threshold for determining cloud cutoff
        Equation 17 (Zhu and Woodcock, 2012)
        Parameters
        ----------
        land_cloud_prob: ndarray
            probability of cloud over land
        pcps: ndarray
            potential cloud pixels
        water: ndarray
            water mask
        Output
        ------
        float:
            land cloud threshold
        """
        # eq 12
        clearsky_land = ~(pcps | water)

        # 82.5th percentile of lCloud_Prob(masked by clearsky_land) + LE07_clip_L1TP_039027_20150529_20160902_01_T1_B1.TIF.2
        cloud_prob = land_cloud_prob.copy()
        cloud_prob[~clearsky_land] = np.nan
        cloud_prob[~self.mask] = np.nan

        # eq 17
        th_const = 0.2
        return np.nanpercentile(cloud_prob, 82.5) + th_const

    def potential_cloud_layer(self, pcp, water, tlow, land_cloud_prob, land_threshold,
                              water_cloud_prob, water_threshold=0.5):
        """Final step of determining potential cloud layer
        Equation 18 (Zhu and Woodcock, 2012)
        
        Saturation (green or red) test is not in the algorithm
        
        Parameters
        ----------
        pcps: ndarray
            potential cloud pixels
        water: ndarray
            water mask
        tirs1: ndarray
        tlow: float
            low percentile of land temperature
        land_cloud_prob: ndarray
            probability of cloud over land
        land_threshold: float
            cutoff for cloud over land
        water_cloud_prob: ndarray
            probability of cloud over water
        water_threshold: float
            cutoff for cloud over water
        Output
        ------
        ndarray:
            potential cloud layer, boolean
        """
        # Using pcp and water as mask todo
        # change water threshold to dynamic, line 132 in Zhu, 2015 todo
        part1 = (pcp & water & (water_cloud_prob > water_threshold))
        part2 = (pcp & ~water & (land_cloud_prob > land_threshold))
        temptest = self.tirs1 < (tlow - 35)  # 35degrees C colder

        if self.sat in ['LT5', 'LE7']:
            saturation = self.blue_saturated | self.green_saturated | self.red_saturated

            return part1 | part2 | temptest | saturation

        else:
            return part1 | part2 | temptest

    def potential_cloud_shadow_layer(self, water):
        """Find low NIR/SWIR1 that is not classified as water
        This differs from the Zhu Woodcock algorithm
        but produces decent results without requiring a flood-fill
        Parameters
        ----------
        nir: ndarray
        swir1: ndarray
        water: ndarray
        Output
        ------
        ndarray
            boolean, potential cloud shadows
        """
        return (self.nir < 0.10) & (self.swir1 < 0.10) & ~water

    def potential_snow_layer(self):
        """Spectral test to determine potential snow
        Uses the 9.85C (283K) threshold defined in Zhu, Woodcock 2015
        Parameters
        ----------
        ndsi: ndarray
        green: ndarray
        nir: ndarray
        tirs1: ndarray
        Output
        ------
        ndarray:
            boolean, True is potential snow
        """
        return (self.ndsi > 0.15) & (self.tirs1 < 9.85) & (self.nir > 0.11) & (self.green > 0.1)

    def cloud_mask(self, min_filter=(3, 3), max_filter=(10, 10), combined=False, cloud_and_shadow=False):
        """Calculate the potential cloud layer from source data
        *This is the high level function which ties together all
        the equations for generating potential clouds*
        Parameters
        ----------
        blue: ndarray
        green: ndarray
        red: ndarray
        nir: ndarray
        swir1: ndarray
        swir2: ndarray
        cirrus: ndarray
        tirs1: ndarray
        min_filter: 2-element tuple, default=(3,3)
            Defines the window for the minimum_filter, for removing outliers
        max_filter: 2-element tuple, default=(21, 21)
            Defines the window for the maximum_filter, for "buffering" the edges
        combined: make a boolean array masking all (cloud, shadow, water)
        Output
        ------
        ndarray, boolean:
            potential cloud layer; True = cloud
        ndarray, boolean
            potential cloud shadow layer; True = cloud shadow
            :param cloud_and_shadow:
        """
        # logger.info("Running initial testsr")
        whiteness = self.whiteness_index()
        water = self.water_test()

        # First pass, potential clouds
        pcps = self.potential_cloud_pixels()

        if self.sat == 'LC8':
            cirrus_prob = self.cirrus / 0.04
        else:
            cirrus_prob = 0.0

        # Clouds over water
        wtp = self.water_temp_prob()
        bp = self.brightness_prob()
        water_cloud_prob = (wtp * bp) + cirrus_prob
        wthreshold = 0.5

        # Clouds over land
        tlow, thigh = self.temp_land(pcps, water)
        ltp = self.land_temp_prob(tlow, thigh)
        vp = self.variability_prob(whiteness)
        land_cloud_prob = (ltp * vp) + cirrus_prob
        lthreshold = self.land_threshold(land_cloud_prob, pcps, water)

        # logger.info("Calculate potential clouds")
        pcloud = self.potential_cloud_layer(
            pcps, water, tlow,
            land_cloud_prob, lthreshold,
            water_cloud_prob, wthreshold)

        # Ignoring snow for now as it exhibits many false positives and negatives
        # when used as a binary mask
        # psnow = potential_snow_layer(ndsi, green, nir, tirs1)
        # pcloud = pcloud & ~psnow

        # logger.info("Calculate potential cloud shadows")
        pshadow = self.potential_cloud_shadow_layer(water)

        # The remainder of the algorithm differs significantly from Fmask
        # In an attempt to make a more visually appealling cloud mask
        # with fewer inclusions and more broad shapes

        if min_filter:
            # Remove outliers
            # logger.info("Remove outliers with minimum filter")

            from scipy.ndimage.filters import minimum_filter
            from scipy.ndimage.morphology import distance_transform_edt

            # remove cloud outliers by nibbling the edges
            pcloud = minimum_filter(pcloud, size=min_filter)

            # crude, just look x pixels away for potential cloud pixels
            dist = distance_transform_edt(~pcloud)
            pixel_radius = 100.0
            pshadow = (dist < pixel_radius) & pshadow

            # remove cloud shadow outliers
            pshadow = minimum_filter(pshadow, size=min_filter)

        if max_filter:
            # grow around the edges
            # logger.info("Buffer edges with maximum filter")

            from scipy.ndimage.filters import maximum_filter

            pcloud = maximum_filter(pcloud, size=max_filter)
            pshadow = maximum_filter(pshadow, size=max_filter)

        # mystery, save pcloud here, shows no nan in qgis, save later, shows nan
        # outfile = '/data01/images/sandbox/pcloud.tif'
        # georeference = self.sat_image.rasterio_geometry
        # array = pcloud
        # array = array.reshape(1, array.shape[LE07_clip_L1TP_039027_20150529_20160902_01_T1_B1.TIF], array.shape[1])
        # array = np.array(array, dtype=georeference['dtype'])
        # with rasterio.open(outfile, 'w', **georeference) as dst:
        #     dst.write(array)
        # mystery test
        if combined:
            return pcloud | pshadow | water

        if cloud_and_shadow:
            return pcloud | pshadow

        return pcloud, pshadow, water

    def save_array(self, array, outfile):

        print('Writing {}'.format(outfile))
        georeference = self.image.rasterio_geometry

        if array.dtype == bool:
            georeference['dtype'] = rasterio.uint8
            array = array.reshape(1, array.shape[0], array.shape[1])
            with rasterio.open(outfile, 'w', **georeference) as dst:
                dst.write(array.astype(rasterio.uint8))

        else:
            georeference['dtype'] = array.dtype
            array = array.reshape(1, array.shape[0], array.shape[1])
            with rasterio.open(outfile, 'w', **georeference) as dst:
                dst.write(array.astype(georeference['dtype']))

        return None

    @staticmethod
    def gdal_nodata_mask(pcl, pcsl, tirs_arr):
        """
        Given a boolean potential cloud layer,
        a potential cloud shadow layer and a thermal band
        Calculate the GDAL-style uint8 mask
        """
        tirs_mask = np.isnan(tirs_arr) | (tirs_arr == 0)
        return ((~(pcl | pcsl | tirs_mask)) * 255).astype('uint8')

    @staticmethod
    def _divide_zero(a, b, replace=0):
        with np.errstate(divide='ignore', invalid='ignore'):
            c = np.true_divide(a, b)
            c[c == np.inf] = replace
            c = np.nan_to_num(c)
            return c
        return potential_cloud

    @staticmethod
    def _counts(arr):
        return np.count_nonzero(~arr), np.count_nonzero(arr)

# ========================= EOF ====================================================================
